
/*
 * DUDISCDEV.C
 *
 * DUDISCDEV -r rep -i interface -d ttydevice -s speed [-e expect] [-t encap]
 *
 * -r <rep>		0 to report failures, 1 to not report failures
 * -i <interface>	interface base, default is 'sl'
 * -d <ttydevice>	tty device, default is '/dev/cua0'
 * -s <speed>		baud rate to set the port to, default 38400
 * -e <script>		script file to run after tty modes are set (optional)
 *
 *			The script file is responsible for connecting to 
 *			the remote side.  It must return with an exit code
 *			of 2 for DUDISCDEV to continue.
 *
 * -t <encap>		encapsulation (optional).  The default is normally
 *			uncompressed slip.   0 is uncompressed slip, 1 is
 *			compressed slip.  NOTE!  NET2E ALPHA 4 OR BETTER,
 *			I do not know how older networking code deals with
 *			encapsulation.
 *
 * DUDISCDEV returns the detached monitor program's pid which the master
 * script can monitor.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <termios.h>

void eprintf(const char *ctl, ...);

short ReportFailures = 1;

main(ac, av)
char *av[];
{
    char *ifcName = "sl";
    char *devName = "/dev/cua0";
    char *expectFile = NULL;
    int speed = 38400;
    int i;
    int ifcNo = 0;
    int fd;
    int ttyFd;
    int pid;
    int encap = -1;

    for (i = 1; i < ac; ++i) {
    	char *ptr = av[i];

    	if (*ptr != '-') {
    	    if (ReportFailures)
		printf("0:unexpected argument: %s\n", ptr);
    	    exit(0);
    	}
    	ptr += 2;
    	switch(ptr[-1]) {
    	case 'r':
    	    ReportFailures = !strtol((*ptr ? ptr : av[++i]), NULL, 0);
    	    break;
    	case 't':
    	    encap = strtol((*ptr ? ptr : av[++i]), NULL, 0);
    	    break;
    	case 'i':
    	    ifcName = (*ptr) ? ptr : av[++i];
    	    break;
    	case 'd':
    	    devName = (*ptr) ? ptr : av[++i];
    	    break;
    	case 's':
    	    // can be 0, indicating that the speed should not be changed
    	    //
    	    speed = strtol((*ptr ? ptr : av[++i]), NULL, 0);
    	    break;
    	case 'e':
    	    expectFile = (*ptr) ? ptr : av[++i];
    	    break;
	default:
    	    if (ReportFailures)
		eprintf("unexpected option\n", ptr - 2);
	    printf("0\n");
	    exit(0);
	}
    }

    /*
     * attempt to open tty
     */

    fd = open(devName, O_RDWR);
    if (fd <= 0) {
        if (ReportFailures)
	    eprintf("couldn't open %s\n", devName);
	printf("0\n");
        exit(0);
    }

    /*
     * fork, return child pid
     */

    signal(SIGCHLD, SIG_IGN);

    pid = fork();
    if (pid > 0) {
    	printf("%d\n", pid);
    	exit(0);
    }
    if (pid < 0) {
        if (ReportFailures)
	    eprintf("couldn't fork\n");
	printf("0\n");
    	exit(0);
    }
    fclose(stdin);
    fclose(stdout);
/*    fclose(stderr); */

    /*
     * CHILD
     *
     * - disconnect from tty, open new tty to make controlling terminal
     */

    {
	int ttyfd = open("/dev/tty", O_RDWR);
	if (ttyfd >= 0) {
	    ioctl(ttyfd, TIOCNOTTY, 0);
	    close(ttyfd);
	}
	ttyfd = open(devName, O_RDWR);
	if (ttyfd < 0) {
	    if (ReportFailures)
		eprintf("couldn't reopen %s\n", devName);
	    exit(0);
	}
	close(fd);
	if (ttyfd != 0)
	    dup2(ttyfd, 0);
	if (ttyfd != 1)
	    dup2(ttyfd, 1);
/*
	if (ttyfd != 2)
	    dup2(ttyfd, 2);
*/
	if (ttyfd > 2)
	    close(ttyfd);
    }

    /*
     * terminal modes
     */

    {
        struct termios tios;
        int baud = tios.c_cflag & CBAUD;

        switch(speed) {
        case 0:
            break;
	case 50:
	    baud = B50;
	    break;
	case 75:
	    baud = B75;
	    break;
	case 110:
	    baud = B110;
	    break;
	case 134:
	    baud = B134;
	    break;
	case 150:
	    baud = B150;
	    break;
	case 200:
	    baud = B200;
	    break;
	case 300:
	    baud = B300;
	    break;
	case 600:
	    baud = B600;
	    break;
	case 1200:
	    baud = B1200;
	    break;
	case 1800:
	    baud = B1800;
	    break;
	case 2400:
	    baud = B2400;
	    break;
	case 4800:
	    baud = B4800;
	    break;
	case 9600:
	    baud = B9600;
	    break;
	case 19200:
	    baud = B19200;
	    break;
	case 38400:
	    baud = B38400;
	    break;
	default:
	    if (ReportFailures)
		eprintf("bad baud baud rate\n");
	    exit(0);
	}


        if (ioctl(0, TCGETS, &tios) != 0) {
            perror("TCGETS");
            exit(0);
        }
        bzero(tios.c_cc, sizeof(tios.c_cc));
        tios.c_iflag = IGNBRK|IGNPAR;
        tios.c_oflag = 0;
        tios.c_cflag = CRTSCTS|CS8|CREAD|HUPCL|baud;
        tios.c_lflag = 0;

        if (ioctl(0, TCSETS, &tios) != 0) {
            perror("TCSETS");
            exit(0);
        }
    }

    /*
     * I/O error (nothing connected?)
     */

    write(1, "\r", 1);
    sleep(1);
    if (write(1, "\r", 1) != 1) {
	if (ReportFailures)
	    eprintf("I/O error on port\n");
        exit(0);
    }

    /*
     * run expect script
     */

    if (expectFile) {
        int n;

        n = system(expectFile);

	if (n >= 256)	// NeXT 3.0 hack
	    n >>= 8;
	if (n != 2) {
	    if (ReportFailures)
		eprintf("expect script failed\n");
	    exit(0);
	}
    }

    /*
     * check carrier
     */

    sleep(1);

    {
	int n;

	ioctl(0, TIOCMGET, &n);
	if ((n & TIOCM_CAR) == 0) {
	    if (ReportFailures)
		eprintf("expected carrier\n");
	    exit(0);
	}
    }

    /*
     * line discipline
     */

    {
    	int n = N_SLIP;

	if ((ifcNo = ioctl(0, TIOCSETD, &n)) < 0) {
	    perror("set line discipline");
	    exit(0);
	}
	if (ioctl(0, TIOCGETD, &n) || n != N_SLIP) {
	    perror("verify line discipline");
	    exit(0);
	}
	if (encap >= 0) {
	    n = encap;
	    if (ioctl(0, SIOCSIFENCAP, &n) != 0) {
		if (ReportFailures)
		    eprintf("error setting encapsulation\n");
	    }
	}
    }

    sprintf(av[0], "%s%d", ifcName, ifcNo);

    /*
     * wait for carrier lost
     */

    for (;;) {
	int n;

	ioctl(0, TIOCMGET, &n);
	if ((n & TIOCM_CAR) == 0)
	    break;
	sleep(5);
    }

    eprintf("carrier lost\n");

    /*
     * carrier lost, switch back to TTY line disc.  Must close descriptors
     * first
     */

    {
        int n = N_TTY;

        close(0);
        close(1);
        close(2);
        ioctl(0, TIOCSETD, &n);
    }
    exit(0);
}

void
eprintf(const char *ctl, ...)
{
    va_list va;
    char buf[64];
    time_t t = time(NULL);
    struct tm *tm = localtime(&t);

    strftime(buf, sizeof(buf), "%a %b %d %X %Y", tm);
    fprintf(stderr, "%s ", buf);
    va_start(va, ctl);
    vfprintf(stderr, ctl, va);
    va_end(va);
}

